/*
 * A program that prints out an ensemble energy of nucleic acid input.
 *
 * (c) 2010  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "EnsembleEnergy_Interface.h"

///////////////////////////////////////////////////////////////////////////////
// Constructor.
///////////////////////////////////////////////////////////////////////////////
EnsembleEnergy_Interface::EnsembleEnergy_Interface() {

  // Initialize the type description.
  calcType = "Ensemble energy calculation";

  // Initialize the sequence flag to false.
  isSequence = false;

  // Initialize the silent flag to false.
  isSilent = false;

}

///////////////////////////////////////////////////////////////////////////////
// Parse the command line arguments.
///////////////////////////////////////////////////////////////////////////////
bool EnsembleEnergy_Interface::parse( int argc, char** argv ) {

  // Initialize array of flags without parameters.
  const char* flagsNoParams[] = {
    "-d", "-D", "--DNA",
    "-s", "-S", "--silent",
    "--sequence"
  };

  // Initialize usage string.
  string usageString =
    "USAGE: EnsembleEnergy <input file> [options]";

  // Initialize required parameters string.
  string inputHelp = "\
<input file>\n\
     The name of the input file. This file can be in two formats:\n\
     1. Partition function save file (holds probability data)\n\
     2. Sequence file (holds raw sequence: .seq or .fasta).\n\
        Note that in order to use a sequence file, the \"--sequence\" flag\n\
        must be specified (see \"--sequence\" below).\
";

  string paramString =
    inputHelp;

  // Initialize required parameters string.
  string dnaHelp = "\
-d, -D, --DNA\n\
     This flag only matters if the input file is a sequence file and has\n\
     been specified as such (see \"--sequence\" below).\n\
     Specify that the sequence is DNA, and DNA parameters are to be used.\n\
     Default is to use RNA parameters.\
";

  string helpLabel = "\
-h, -H, --help\n\
     Display the usage details message.\
";

  string seqLabel = "\
--sequence\n\
     Identify the input file format as a sequence file.\
";

  string silentLabel = "\
-s, -S, --silent\n\
     Suppress all progress messages except the final ensemble energy result.\n\
     Note that this does NOT suppress errors.\
";

  string flagsNoParamsString =
    dnaHelp + "\n\n" + helpLabel + "\n\n" + seqLabel + "\n\n" + silentLabel;

  // Initialize flags with parameters string.
  string flagsParamsString =
    "";

  // Initialize parser and check the command line.
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 1,
						   7, flagsNoParams,
						   0, NULL );
  parser->setUsageStrings( usageString.c_str(), paramString.c_str(),
                           flagsNoParamsString.c_str(),
			   flagsParamsString.c_str() );
  bool noError = parser->checkLine();

  // If command line structure is OK, proceed with argument retrieval, parsing.
  if( noError ) {

    // Get required parameters.
    input = parser->getParameter( 1 );

    // Set non-parameterized flag(s). The RNA/DNA flag is set only if a
    // sequence flag is present.
    isSequence = parser->contains( "--sequence" );
    if( isSequence ) { isRNA = !parser->containsInGroup( "-d -D --DNA" ); }
    isSilent = parser->containsInGroup( "-s -S --silent" );
  }

  // Delete the parser and return the error status.
  delete parser;
  return noError;

}

///////////////////////////////////////////////////////////////////////////////
// Run ensemble energy determination.
///////////////////////////////////////////////////////////////////////////////
void EnsembleEnergy_Interface::run() {

  // Create a variable to handle errors.
  int error = 0;

  /*
   * Use the constructor for RNA that specifies a filename.
   * If the input file is a pfs file, specify type = 3 (pfs file).
   * If the input file is a sequence file, specify type = 2 (sequence file).
   *
   * After construction of the strand, create the error checker which monitors
   * the strand for errors. Then, check for errors with the isErrorStatus
   * function, which returns 0 if no error occurs.
   * Throughout, the calculation proceeds as long as error = 0.
   */
  if( !isSilent ) { cout << "Initializing nucleic acids..." << flush; }
  int type = ( !isSequence ) ? 3 : 2;
  RNA* strand = new RNA( input.c_str(), type, isRNA );
  ErrorChecker<RNA>* checker = new ErrorChecker<RNA>( strand );
  error = checker->isErrorStatus();
  if( ( !isSilent ) && ( error == 0 ) ) { cout << "done." << endl; }

  /*
   * If the input file is a sequence file, calculate the partition function
   * first before calculating EnsembleEnergy_Interface sampling.
   */
  if( error == 0 && isSequence ) {

    // Print a message saying partition function has started, if suppression
    // of messages is not enabled.
    if( !isSilent ) { cout << "Calculating partition function..." << flush; }

    // Run the partition function.
    int partError = strand->PartitionFunction();

    // Check the error status.
    error = checker->isErrorStatus( partError );

    // If no error occurred and suppression of messages is not enabled, print
    // out a message saying partition function is done.
    if( ( !isSilent ) && ( error == 0 ) ) { cout << "done." << endl; }
  }

  /*
   * Calculate the ensemble energy using the GetEnsembleEnergy method.
   * After the main calculation is complete, use the error checker's
   * isErrorStatus method to check for errors.
   */
  if( error == 0 ) {

    // Show a message that the main calculation has started, if suppression of
    // messages is not enabled.
    if( !isSilent ) { cout << "Calculating ensemble energy..." << flush; }

    // Calculate ensemble energy and check for errors.
    double energy = strand->GetEnsembleEnergy();
    error = checker->isErrorStatus();

    // If no error occurred and suppression of messages is not enabled, print
    // message that main calculation is done.
    if( ( !isSilent ) && ( error == 0 ) ) { cout << "done." << endl << endl; }

    // If no error occurred, print out the ensemble energy.
    cout << "Ensemble energy for " << input << ": " << energy << " kcal/mol"
	 << endl << endl;
  }

  // Delete the error checker and data structure.
  delete checker;
  delete strand;

  // Print confirmation of run finishing.
  if( error == 0 ) {
    if( !isSilent ) { cout << calcType << " complete." << endl; }
  } else { cerr << calcType << " complete with errors." << endl; }

}

///////////////////////////////////////////////////////////////////////////////
// Main method to run the program.
///////////////////////////////////////////////////////////////////////////////
int main( int argc, char* argv[] ) {

  EnsembleEnergy_Interface* runner = new EnsembleEnergy_Interface();
  bool parseable = runner->parse( argc, argv );
  if( parseable == true ) { runner->run(); }
  delete runner;
  return 0;

}
